import os
from sass import compile
from django.conf import settings
from django.middleware import cache

def locate(filename):
    for path, dirs, files in os.walk(os.getcwd(), followlinks=True):
        for f in files:
            if f == filename:
                yield os.path.join(path, filename)

class FetchFromCacheMiddleware(cache.FetchFromCacheMiddleware):
    '''Minor change to the original middleware that prevents caching of
    requests that have a `sessionid` cookie. This should be the
    Django default, IMHO.

    '''
    def process_request(self, request):
        if 'sessionid' not in request.COOKIES:
            return super().process_request(request)

class SassMiddleware:
    '''Simple SASS middleware that intercepts requests for .css files and
    tries to compile the corresponding SCSS file.

    '''
    def __init__(self, get_response):
        self.get_response = get_response

    def __call__(self, request):
        if settings.DEBUG and request.path.endswith('.css'):
            css_file = request.path.rsplit('/',1)[1]
            sass_file = css_file[:-4]
            sass_paths = locate(sass_file)
            for sass_path in list(sass_paths):
                if os.path.exists(sass_path):
                    css_path = sass_path + '.css'
                    map_path = css_path + '.map'
                    css = compile(filename=sass_path, output_style='nested')
                    css, mapping = compile(filename=sass_path, source_map_filename=map_path)
                    with open(css_path, 'w') as f:
                        f.write(css)
                    with open(map_path, 'w') as f:
                        f.write(mapping)

        response = self.get_response(request)
        return response



import os, sys, random, string

DEBUG = 'runserver' in sys.argv
PROJECT_NAME = os.path.basename(os.path.dirname(os.path.abspath(__file__)))
KEYFILE = f'/tmp/{PROJECT_NAME}.secret'
ADMINS = [('JJ Vens', 'jj@rtts.eu')]
DEFAULT_FROM_EMAIL = 'noreply@rtts.eu'
DEFAULT_TO_EMAIL = ['jj@rtts.eu']
ALLOWED_HOSTS = ['*']
ROOT_URLCONF = PROJECT_NAME + '.urls'
WSGI_APPLICATION = PROJECT_NAME + '.wsgi.application'
LANGUAGE_CODE = 'nl'
TIME_ZONE = 'Europe/Amsterdam'
USE_I18N = True
USE_L10N = True
USE_TZ = True
STATIC_URL = '/static/'
STATIC_ROOT = '/srv/' + PROJECT_NAME + '/static'
MEDIA_URL = '/media/'
MEDIA_ROOT = '/srv/' + PROJECT_NAME + '/media'
LOGIN_REDIRECT_URL = '/'
LOGOUT_REDIRECT_URL = '/'

if DEBUG:
    EMAIL_BACKEND = 'django.core.mail.backends.console.EmailBackend'
    CACHE_MIDDLEWARE_SECONDS = 0

try:
    with open(KEYFILE) as f:
        SECRET_KEY = f.read()
except IOError:
    SECRET_KEY = ''.join(random.choice(string.printable) for x in range(50))
    with open(KEYFILE, 'w') as f:
        f.write(SECRET_KEY)

INSTALLED_APPS = [
    PROJECT_NAME,
    'django.contrib.admin',
    'django.contrib.auth',
    'django.contrib.contenttypes',
    'django.contrib.sessions',
    'django.contrib.messages',
    'cms',
    'embed_video',
    'easy_thumbnails',
    'django_extensions',
]
if not DEBUG:
    INSTALLED_APPS += ['django.contrib.staticfiles']

MIDDLEWARE = [
    'django.middleware.cache.UpdateCacheMiddleware',
    'cms.middleware.SassMiddleware',
    'django.middleware.security.SecurityMiddleware',
    'django.contrib.sessions.middleware.SessionMiddleware',
    'django.middleware.common.CommonMiddleware',
    'django.middleware.csrf.CsrfViewMiddleware',
    'django.contrib.auth.middleware.AuthenticationMiddleware',
    'django.contrib.messages.middleware.MessageMiddleware',
    'django.middleware.clickjacking.XFrameOptionsMiddleware',
    'cms.middleware.FetchFromCacheMiddleware',
]

TEMPLATES = [
    {
        'BACKEND': 'django.template.backends.django.DjangoTemplates',
        'DIRS': [],
        'APP_DIRS': True,
        'OPTIONS': {
            'context_processors': [
                'django.template.context_processors.debug',
                'django.template.context_processors.request',
                'django.contrib.auth.context_processors.auth',
                'django.contrib.messages.context_processors.messages',
            ],
        },
    },
]

DATABASES = {
    'default': {
        'ENGINE': 'django.db.backends.postgresql',
        'USER': PROJECT_NAME,
        'NAME': PROJECT_NAME,
    }
}
CACHES = {
    'default': {
        'BACKEND': 'django.core.cache.backends.memcached.PyLibMCCache',
        'LOCATION': '127.0.0.1:11211',
        'KEY_PREFIX': PROJECT_NAME,
    }
}



from django.utils.translation import gettext_lazy as _
from cms.views import SectionView, SectionFormView
from cms.decorators import section_view
from cms.forms import ContactForm

@section_view
class Text(SectionView):
    verbose_name = _('Text')
    fields = ['content']
    template_name = 'text.html'

@section_view
class Images(SectionView):
    verbose_name = _('Image(s)')
    fields = ['images']
    template_name = 'images.html'

@section_view
class Video(SectionView):
    verbose_name = _('Video')
    fields = ['video']
    template_name = 'video.html'

@section_view
class Contact(SectionFormView):
    verbose_name = _('Contact')
    fields = []
    form_class = ContactForm
    success_url = '/thanks/'
    template_name = 'contact.html'



import os, argparse, shutil, example
from pip._internal.operations import freeze as pip

def main():
    parser = argparse.ArgumentParser(description='SimpleCMS')
    parser.add_argument('project_name', nargs='?')
    if project_name := parser.parse_args().project_name:
        project_dir = os.path.join(os.getcwd(), project_name)
    else:
        project_name = os.path.basename(os.getcwd())
        project_dir = os.getcwd()
    if input(f'Do you want to create a new project in `{project_dir}` ?\033[1D') in 'Yy':
        create_project(project_name, project_dir)

def create_project(project_name, project_dir):
    os.makedirs(project_dir, exist_ok=True)
    with open(os.path.join(project_dir, 'requirements.txt'), 'w') as f:
        for line in pip.freeze():

            # Central point of failure
            if 'django_simplecms' in line:
                line = 'git+https://github.com/rtts/django-simplecms'

            print(line, file=f)

    shutil.copytree(os.path.dirname(example.__file__), os.path.join(project_dir, project_name), dirs_exist_ok=True)
    with open(os.open(os.path.join(project_dir, 'manage.py'), os.O_CREAT|os.O_WRONLY, 0o755), 'w') as f:
        print('''#!/usr/bin/env python
import os, sys
os.environ.setdefault('DJANGO_SETTINGS_MODULE', '{project_name}.settings')
from django.core.management import execute_from_command_line
execute_from_command_line(sys.argv)''', file=f)
    with open(os.path.join(project_dir, '.gitignore'), 'w') as f:
        print('*.pyc\n__pycache__/', file=f)

    print(f'''
Successfully created project "{project_name}"

Things to do next:
- create a database
- ./manage.py makemigrations
- ./manage.py migrate
- ./manage.py createsuperuser
- ./manage.py runserver
''')

if __name__ == '__main__':
    main()



from django.db import models
from django.utils.translation import gettext_lazy as _
from cms.models import BasePage, BaseSection
from cms.decorators import page_model, section_model

@page_model
class Page(BasePage):
    '''Add custom fields here. Already existing fields: title, slug,
    number, menu

    '''

@section_model
class Section(BaseSection):
    '''Add custom fields here. Already existing fields: title, type,
    number, content, image, video, href

    '''
    page = models.ForeignKey(Page, related_name='sections', on_delete=models.PROTECT)

class SectionImage(models.Model):
    section = models.ForeignKey(Section, related_name='images', on_delete=models.CASCADE)
    image = models.ImageField(_('Image'))

    class Meta:
        ordering = ['pk']



page_class = None
section_class = None
section_types = []
view_per_type = {}

def get_types():
    return section_types

def get_view(section, request):
    return view_per_type[section.type](request)

def get_fields_per_type():
    fields_per_type = {}
    for name, view in view_per_type.items():
        fields_per_type[name] = ['title', 'type', 'number'] + view.fields
    return fields_per_type




from django.urls import path
from .views import PageView, CreatePage, UpdatePage, CreateSection, UpdateSection

app_name = 'cms'

urlpatterns = [
    path('new/', CreatePage.as_view(), name='createpage'),
    path('edit/', UpdatePage.as_view(), kwargs={'slug': ''}, name='updatepage'),
    path('edit/<int:number>/', UpdateSection.as_view(), kwargs={'slug': ''}, name='updatesection'),
    path('<slug:slug>/edit/', UpdatePage.as_view(), name='updatepage'),
    path('<slug:slug>/edit/new/', CreateSection.as_view(), name='createsection'),
    path('<slug:slug>/edit/<int:number>/', UpdateSection.as_view(), name='updatesection'),
    path('', PageView.as_view(), name='page'),
    path('<slug:slug>/', PageView.as_view(), name='page'),
]



from django import forms
from django.conf import settings
from django.db.models import Prefetch
from django.contrib.contenttypes.models import ContentType
from django.core.mail import EmailMessage
from django.utils.translation import gettext_lazy as _

from . import registry

class PageForm(forms.ModelForm):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.label_suffix = ''

        self.formsets = [forms.inlineformset_factory(
            parent_model=registry.page_class,
            model=registry.section_class,
            form=SectionForm,
            extra=1,
        )(
            data=self.data if self.is_bound else None,
            files=self.files if self.is_bound else None,
            instance=self.instance,
        )]
        self.formsets[0][0].empty_permitted = True

    def is_valid(self):
        return super().is_valid() and self.formsets[0].is_valid()

    def has_changed(self):
        return super().has_changed() or self.formsets[0].has_changed()

    def clean(self):
        super().clean()
        if not self.formsets[0].is_valid():
            self.add_error(None, repr(self.formsets[0].errors))

    def save(self, *args, **kwargs):
        page = super().save()
        formset = self.formsets[0]
        formset.instance = page
        formset.save()
        return page

    class Meta:
        model = registry.page_class
        fields = '__all__'

class SectionForm(forms.ModelForm):
    type = forms.ChoiceField()

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.label_suffix = ''
        self.fields['DELETE'] = forms.BooleanField(label=_('Delete'), required=False)
        self.fields['type'].choices = registry.get_types()
        self.fields['type'].initial = registry.get_types()[0][0]

        # Populate the 'formsets' attribute if the Section was
        # extendend with one_to_many fields
        self.formsets = []
        for field in self.instance._meta.get_fields():
            if field.one_to_many:
                formset = forms.inlineformset_factory(
                    parent_model=registry.section_class,
                    model=field.related_model,
                    fields='__all__',
                    extra=1,
                )(
                    instance=self.instance,
                    data=self.data if self.is_bound else None,
                    files=self.files if self.is_bound else None,
                    prefix=f'{self.prefix}-{field.name}',
                    form_kwargs={'label_suffix': self.label_suffix},
                )
                formset.name = field.name
                self.formsets.append(formset)

    def is_valid(self):
        result = super().is_valid()
        for formset in self.formsets:
            result = result and formset.is_valid() # AND
        return result

    def has_changed(self):
        result = super().has_changed()
        for formset in self.formsets:
            result = result or formset.has_changed() # OR
        return result

    def save(self, commit=True):
        section = super().save(commit=commit)

        if self.cleaned_data['DELETE']:
            section.delete()
            if section.page.slug and not section.page.sections.exists():
                section.page.delete()
            return
        elif commit:
            section.save()

        for formset in self.formsets:
            formset.save(commit=commit)

        return section

    class Meta:
        model = registry.section_class
        exclude = ['page']

class ContactForm(forms.Form):
    sender = forms.EmailField(label=_('Your email address'))
    spam_protection = forms.CharField(label=_('Your message'), widget=forms.Textarea())
    message = forms.CharField(label=_('Your message'), widget=forms.Textarea(), initial='Hi there!')

    def save(self):
        body = self.cleaned_data.get('spam_protection')
        if len(body.split()) < 7:
            return
        spamcheck = self.cleaned_data.get('message')
        if spamcheck != 'Hi there!':
            return

        email = EmailMessage(
            to = settings.DEFAULT_TO_EMAIL,
            body = body,
            subject = _('Contact form'),
            headers = {'Reply-To': self.cleaned_data.get('sender')},
        )
        email.send()



import json

from django.shortcuts import redirect
from django.views.generic import base, detail, edit
from django.utils.decorators import method_decorator
from django.views.decorators.cache import never_cache
from django.contrib.auth.mixins import UserPassesTestMixin
from django.http import Http404, HttpResponse, HttpResponseRedirect, HttpResponseBadRequest

from . import registry
from .forms import PageForm, SectionForm

class SectionView:
    '''Generic section view'''
    template_name = 'cms/sections/section.html'

    def __init__(self, request):
        self.request = request

    def get_context_data(self, **kwargs):
        return kwargs

class SectionFormView(SectionView):
    '''Generic section with associated form'''
    form_class = None
    success_url = None

    def get_context_data(self, **kwargs):
        if 'form' not in kwargs:
            kwargs['form'] = self.get_form()
        return kwargs

    def form_valid(self, form):
        form.save()
        return HttpResponseRedirect(self.success_url)

    def get_form_kwargs(self):
        return {}

    def get_form(self, method='get'):
        form_class = self.form_class
        kwargs = self.get_form_kwargs()
        if method == 'post':
            kwargs.update({
                'data': self.request.POST,
                'files': self.request.FILES,
            })
        return form_class(**kwargs)

class PageView(detail.DetailView):
    '''View of a page with heterogeneous sections'''
    model = registry.page_class
    template_name = 'cms/page.html'

    def setup(self, *args, slug='', **kwargs):
        '''Supply a default argument for slug'''
        super().setup(*args, slug=slug, **kwargs)

    def get(self, request, *args, **kwargs):
        '''Instantiate section views and render final response'''
        try:
            page = self.object = self.get_object()
        except Http404:
            app_label = registry.page_class._meta.app_label
            model_name = registry.page_class._meta.model_name
            if self.kwargs['slug'] == '':
                page = registry.page_class(title='Homepage', slug='')
                page.save()
                self.object = page

            # Special case: Don't serve 404 pages to authorized users,
            # but redirect to the edit page form with the slug pre-filled
            elif (self.request.user.has_perm(f'{app_label}.change_{model_name}')):
                return redirect('cms:updatepage', slug=self.kwargs['slug'])
            else:
                raise
        context = self.get_context_data(**kwargs)
        sections = page.sections.all()
        context.update({
            'page': page,
            'sections': sections,
        })
        return self.render_to_response(context)

    def post(self, request, **kwargs):
        '''Call the post() method of the correct section view'''
        try:
             pk = int(self.request.POST.get('section'))
        except:
            return HttpResponseBadRequest()

        page = self.object = self.get_object()
        context = self.get_context_data(**kwargs)
        sections = page.sections.all()
        for section in sections:
            if section.pk == pk:
                view = registry.get_view(section, request)
                form = view.get_form(method='post')
                if form.is_valid():
                    return view.form_valid(form)
                section.invalid_form = form

        context.update({
            'page': page,
            'sections': sections,
        })
        return self.render_to_response(context)

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        pages = registry.page_class.objects.filter(menu=True)
        context.update({
            'pages': pages,
        })
        return context

@method_decorator(never_cache, name='dispatch')
class EditPage(UserPassesTestMixin, edit.ModelFormMixin, base.TemplateResponseMixin, base.View):
    '''Base view with nested forms for editing the page and all its sections'''
    model = registry.page_class
    form_class = PageForm
    template_name = 'cms/edit.html'

    def test_func(self):
        '''Only allow users with the correct permissions'''
        app_label = registry.page_class._meta.app_label
        model_name = registry.page_class._meta.model_name
        return self.request.user.has_perm(f'{app_label}.change_{model_name}')

    def get_form_kwargs(self):
        '''Set the default slug to the current URL for new pages'''
        kwargs = super().get_form_kwargs()
        if 'slug' in self.kwargs:
            kwargs.update({'initial': {'slug': self.kwargs['slug']}})
        return kwargs

    def get_context_data(self, **kwargs):
        '''Populate the fields_per_type dict for use in javascript'''
        context = super().get_context_data(**kwargs)
        context['fields_per_type'] = json.dumps(registry.get_fields_per_type())
        return context

    def get_object(self):
        '''Prevent 404 by serving the new object form'''
        try:
            return super().get_object()
        except Http404:
            return None

    def get(self, *args, **kwargs):
        '''Handle GET requests'''
        self.object = self.get_object()
        return self.render_to_response(self.get_context_data(**kwargs))

    def post(self, *args, **kwargs):
        '''Handle POST requests'''
        self.object = self.get_object()
        form = self.get_form()

        if form.is_valid():
            page = form.save()
            if page:
                return HttpResponseRedirect(page.get_absolute_url())
            return HttpResponseRedirect('/')
        return self.render_to_response(self.get_context_data(form=form, **kwargs))

class CreatePage(EditPage):
    '''View for creating new pages'''
    def get_object(self):
        return registry.page_class()

class UpdatePage(EditPage):
    '''View for editing existing pages'''

@method_decorator(never_cache, name='dispatch')
class EditSection(UserPassesTestMixin, edit.ModelFormMixin, base.TemplateResponseMixin, base.View):
    model = registry.section_class
    form_class = SectionForm
    template_name = 'cms/edit.html'

    def test_func(self):
        '''Only allow users with the correct permissions'''
        app_label = registry.section_class._meta.app_label
        model_name = registry.section_class._meta.model_name
        return self.request.user.has_perm(f'{app_label}.change_{model_name}')

    def get_form_kwargs(self):
        kwargs = super().get_form_kwargs()
        kwargs.update({
            'prefix': 'section',
        })
        return kwargs

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context['fields_per_type'] = json.dumps(registry.get_fields_per_type())
        return context

    def get_object(self, queryset=None):
        try:
            self.page = registry.page_class.objects.get(slug=self.kwargs['slug'])
        except registry.page_class.DoesNotExist:
            raise Http404()
        return self.get_section()

    def get_section(self):
        try:
            section = self.page.sections.get(number=self.kwargs['number'])
        except self.page.sections.DoesNotExist:
            raise Http404()
        return section

    def get(self, *args, **kwargs):
        self.object = self.get_object()
        return self.render_to_response(self.get_context_data(**kwargs))

    def post(self, *args, **kwargs):
        self.object = self.get_object()
        form = self.get_form()

        if form.is_valid():
            section = form.save()
            if section:
                return HttpResponseRedirect(section.get_absolute_url())
            elif self.page.sections.exists():
                return HttpResponseRedirect(self.page.get_absolute_url())
            else:
                return HttpResponseRedirect('/')
        return self.render_to_response(self.get_context_data(form=form, **kwargs))

class CreateSection(EditSection):
    def get_section(self):
        return registry.section_class(page=self.page)

class UpdateSection(EditSection):
    pass



from markdown import markdown as md

from django import template
from django.shortcuts import reverse
from django.utils.safestring import mark_safe
from django.utils.translation import gettext_lazy as _

from cms import registry

MARKDOWN_EXTENSIONS = ['extra', 'smarty']
register = template.Library()

@register.simple_tag(takes_context=True)
def eval(context, expr):
    '''USE WITH CAUTION!!!

    This template tag runs its argument through Django's templating
    system using the current context, placing all power into the
    hands of the content editors.

    Also, it applies Markdown.

    '''
    result = template.Template(expr).render(context)
    return mark_safe(md(result, extensions=MARKDOWN_EXTENSIONS))

@register.simple_tag(takes_context=True)
def editsection(context, inner):
    '''Renders a simple link to edit the current section'''
    section = context['section']
    user = context['request'].user
    app_label = section._meta.app_label
    model_name = section._meta.model_name
    if user.has_perm(f'{app_label}.change_{model_name}'):
        slug = section.page.slug
        number = section.number
        url = reverse('cms:updatesection', args=[slug, number]) if slug else reverse('cms:updatesection', args=[number])
        return mark_safe(f'<a class="edit section" href="{url}">{inner}</a>')
    return ''

@register.simple_tag(takes_context=True)
def editpage(context, inner):
    '''Renders a simple link to edit the current page'''
    page = context['page']
    user = context['request'].user
    app_label = page._meta.app_label
    model_name = page._meta.model_name
    if user.has_perm(f'{app_label}.change_{model_name}'):
        slug = page.slug
        url = reverse('cms:updatepage', args=[slug]) if slug else reverse('cms:updatepage')
        return mark_safe(f'<a class="edit page" href="{url}">{inner}</a>')
    return ''

@register.tag('include_section')
def do_include(parser, token):
    '''Renders the section with its own context'''
    _, section = token.split_contents()
    return IncludeSectionNode(section)

class IncludeSectionNode(template.Node):
    def __init__(self, section):
        self.section = template.Variable(section)
        self.csrf_token = template.Variable('csrf_token')
        self.request = template.Variable('request')
        self.perms = template.Variable('perms')
        super().__init__()

    def render(self, context):
        section = self.section.resolve(context)
        csrf_token = self.csrf_token.resolve(context)
        request = self.request.resolve(context)
        perms = self.perms.resolve(context)

        view = registry.get_view(section, request)
        initial_context = {
            'csrf_token': csrf_token,
            'section': section,
            'request': request,
            'perms': perms,
        }
        if hasattr(section, 'invalid_form'):
            initial_context['form'] = section.invalid_form

        section_context = view.get_context_data(**initial_context)
        t = context.template.engine.get_template(view.template_name)
        return t.render(template.Context(section_context))



from django.conf import settings
from django.contrib import admin
from django.urls import path, include
from django.conf.urls.static import static
from django.views.generic import RedirectView
from django.contrib.staticfiles.urls import staticfiles_urlpatterns

admin.site.site_header = settings.PROJECT_NAME.capitalize()
admin.site.site_title = settings.PROJECT_NAME.capitalize()
urlpatterns = staticfiles_urlpatterns() + static(settings.MEDIA_URL, document_root=settings.MEDIA_ROOT)
urlpatterns += [
    path('admin/', admin.site.urls),
    path('accounts/', include('django.contrib.auth.urls')),
    path('login/', RedirectView.as_view(url='/accounts/login/')),
    path('logout/', RedirectView.as_view(url='/accounts/logout/')),
    path('', include('cms.urls', namespace='cms')),
]



from django.apps import AppConfig
from django.utils.translation import gettext_lazy as _
from django.utils.module_loading import autodiscover_modules

class CmsConfig(AppConfig):
    name = 'cms'
    verbose_name = _('Content Management System')

    def ready(self):
        autodiscover_modules('views')



#!/usr/bin/env python
from setuptools import setup, find_packages

with open('README.md', 'r') as fh:
    long_description = fh.read()

setup(
    name = 'django-simplecms',
    description = 'Simple Django CMS',
    version = '1.0.3',
    author = 'Jaap Joris Vens',
    author_email = 'jj+cms@rtts.eu',
    url = 'https://github.com/rtts/django-simplecms',
    long_description = long_description,
    long_description_content_type = 'text/markdown',
    packages = find_packages(),
    entry_points = {
        'console_scripts': ['simplecms=cms.__main__:main'],
    },
    include_package_data = True,
    classifiers = [
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)',
        'Operating System :: OS Independent',
    ],
    python_requires = '>=3.8',
    install_requires = [
        'django',
        'django-extensions',
        'django-embed-video',
        'easy-thumbnails',
        'libsass',
        'markdown',
        'psycopg2',
        'pylibmc',
    ],
)



from cms import registry

def page_model(cls):
    '''Decorator to register the Page model'''
    registry.page_class = cls
    return cls

def section_model(cls):
    '''Decorator to register the Section model'''
    registry.section_class = cls
    return cls

def section_view(cls):
    '''Decorator to register a view for a specific section'''
    registry.view_per_type[cls.__name__.lower()] = cls
    registry.section_types.append((cls.__name__.lower(), cls.verbose_name))
    return cls



from django.db import models
from django.urls import reverse
from django.utils.text import slugify
from django.forms import TextInput, Select
from django.utils.translation import gettext_lazy as _
from django.core.exceptions import ImproperlyConfigured

from embed_video.fields import EmbedVideoField

class VarCharField(models.TextField):
    '''Variable width CharField'''
    def formfield(self, **kwargs):
        kwargs.update({'widget': TextInput})
        return super().formfield(**kwargs)

class Numbered:
    '''Mixin for numbered models. Overrides the save() method to
    automatically renumber all instances returned by
    number_with_respect_to()

    '''
    def number_with_respect_to(self):
        return self.__class__.objects.all()

    def get_field_name(self):
        return self.__class__._meta.ordering[-1].lstrip('-')

    def _renumber(self):
        '''Renumbers the queryset while preserving the instance's number'''

        queryset = self.number_with_respect_to()
        field_name = self.get_field_name()
        this_nr = getattr(self, field_name)
        if this_nr is None:
            this_nr = len(queryset) + 1

        # The algorithm: loop over the queryset and set each object's
        # number to the counter. When an object's number equals the
        # number of this instance, set this instance's number to the
        # counter, increment the counter by 1, and finish the loop
        counter = 1
        inserted = False
        for other in queryset.exclude(pk=self.pk):
            other_nr = getattr(other, field_name)
            if counter >= this_nr and not inserted:
                setattr(self, field_name, counter)
                inserted = True
                counter += 1
            if other_nr != counter:
                setattr(other, field_name, counter)
                super(Numbered, other).save()
            counter += 1
        if not inserted:
            setattr(self, field_name, counter)

    def save(self, *args, **kwargs):
        self._renumber()
        super().save(*args, **kwargs)

    def delete(self, *args, **kwargs):
        setattr(self, self.get_field_name(), 9999) # hack
        self._renumber()
        super().delete(*args, **kwargs)

class BasePage(Numbered, models.Model):
    '''Abstract base model for pages'''
    title = VarCharField(_('page'))
    slug = models.SlugField(_('slug'), blank=True, unique=True)
    number = models.PositiveIntegerField(_('number'), blank=True)
    menu = models.BooleanField(_('visible in menu'), default=True)

    def __str__(self):
        if not self.pk:
            return str(_('New page'))
        return self.title

    def get_absolute_url(self):
        if self.slug:
            return reverse('cms:page', args=[self.slug])
        return reverse('cms:page')

    class Meta:
        abstract = True
        verbose_name = _('Page')
        verbose_name_plural = _('Pages')
        ordering = ['number']

class BaseSection(Numbered, models.Model):
    '''Abstract base model for sections'''
    TYPES = []
    title = VarCharField(_('section'))
    type = VarCharField(_('type'))
    number = models.PositiveIntegerField(_('number'), blank=True)
    content = models.TextField(_('content'), blank=True)
    image = models.ImageField(_('image'), blank=True)
    video = EmbedVideoField(_('video'), blank=True, help_text=_('Paste a YouTube, Vimeo, or SoundCloud link'))
    href = VarCharField(_('link'), blank=True)

    def number_with_respect_to(self):
        return self.page.sections.all()

    def get_absolute_url(self):
        return self.page.get_absolute_url() + '#' + slugify(self.title)

    def __str__(self):
        if not self.pk:
            return str(_('New section'))
        elif not self.title:
            return str(_('Untitled'))
        else:
            return self.title

    class Meta:
        abstract = True
        verbose_name = _('section')
        verbose_name_plural = _('sections')
        ordering = ['number']
